@@ -34,7 +34,7 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
       else {tf.mod[[j]] <- paste("~",paste(var,collapse="+"))}
       cat("Stage",j,"treatment-free model:",tf.mod[[j]],"\n")
       tf.mod[[j]] <- as.formula(tf.mod[[j]])
-      
+
     }
     for (j in 1:K) {
       var <- c()
@@ -88,7 +88,7 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
   data <- data.frame(data)
   rownames(data) <- seq(1:nrow(data))
   obj$data <- data
-  
+
   obj$blip.mod <- blip.mod
   obj$treat.mod <- treat.mod
   obj$tf.mod <- tf.mod
@@ -125,7 +125,7 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
       if(j==K) remove[new.drop] <- j
       if(j!=K) {
         dropk <- new.drop[!(new.drop %in% drop)]
-        remove[dropk] <- j 
+        remove[dropk] <- j
       }
       drop <- unique(new.drop,drop)
       # if missing, later than stage 1, and requested, calculate IPCW
@@ -138,10 +138,18 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
         H.list <- unique(c(unlist(sapply(blip.mod[(j-1):1],all.vars)),unlist(sapply(tf.mod[(j-1):1],all.vars)),unlist(sapply(treat.mod[(j-1):1],all.vars))))
         H <- data.matrix(data[,which(colnames(data) %in% H.list)])
         H <- H[keep,]
-        pi[which(!apply(is.na(H),1,any))] <- fitted(glm(Miss~.,data=as.data.frame(H)))
-        obj$ipcw[[j]] <- pi
+        pi[which(!apply(is.na(H),1,any))] <- 1 - fitted(glm(Miss~.,data=as.data.frame(H), family = 'binomial'))
+        obj$ipcw[[j]] <- 1/pi
       }
-      keep <- keep[!(keep %in% drop)]
+
+	if (missing == 'ipcw') {
+	keep <- c(1:N)
+      keep <- keep[!(keep %in% new.drop)]				
+	} else {
+      keep <- keep[!(keep %in% drop)]		
+	}
+
+      # only work with those we want to kep at this stage
       pi <- pi[keep]
       A <- A[keep]
       Hpsi <- Hpsi[keep,]
@@ -225,7 +233,7 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
       Hd <- cbind(Hbeta,A*Hpsi)
       if (A.bin == 1) {
         # binary treatment
-        Hw <- cbind(Hbeta,Hpsi*(A - Ahat))
+        Hw <- cbind(Hbeta,Hpsi*(A - Ahat)*pi)
       } else {
         # continuous treatment, requires some fiddly cleanup
         # ***INPUT CLEANUP STARTS
@@ -267,7 +275,7 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
         ifelse(length(p2.list) > 0,Hw <- cbind(Hbeta,Hpsi1*(A-Ahat),Hpsi2*(A2-A2hat)),Hw <- cbind(Hbeta,Hpsi*(A-Ahat)))
       }
       # get estimates
-      est <- solve(t(Hw)%*%Hd) %*% t(Hw) %*% (Y*pi)
+      est <- solve(t(Hw)%*%Hd) %*% t(Hw) %*% (Y)
       # blip parameters
       psi <- est[(dim(Hbeta)[2]+1):(dim(Hbeta)[2]+dim(Hpsi)[2])]
       # non-blip parameters
@@ -339,7 +347,7 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
           }
         }
       }
-      # if linear, then score = (1/s^2) * (A - Ahat)*Halpha 
+      # if linear, then score = (1/s^2) * (A - Ahat)*Halpha
       if (A.bin == 0) {
         score.alpha <- D * Halpha
         score.varsig <- residuals * Hbeta
@@ -396,13 +404,13 @@ DTRreg <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "ge
     k <- j
     while (k < K) {
       k <- k + 1
-      Y.fit <- Y.fit - obj$regret[[k]][which(rownames(obj$regret[[k]])%in% keep)] # correctly use keep indicator
+      Y.fit[which(rownames(obj$regret[[k]])%in% keep)] <- Y.fit[which(rownames(obj$regret[[k]])%in% keep)] - obj$regret[[k]][which(rownames(obj$regret[[k]])%in% keep)] # correctly use keep indicator
     }
     obj$fitted[[j]] <- Y.fit
   }
-  # keep track of which patient are removed at which stage
+  # keep track of which patients are removed at which stage
   obj$remove <- remove
-  # optimal outcome according to blip estimates is current psuedo-Y
+  # optimal outcome according to blip estimates is current pseudo-Y
   obj$opt.Y <- Y
   obj$type <- type
   # standard errors
@@ -605,52 +613,92 @@ coef.DTRreg <- function(object,...) {
 }
 # confidence interval
 #' @export
-confint.DTRreg <- function(object, parm = NULL, level = 0.95, ...) {
-	x <- object
-	K <- x$K
-	obj <- x$psi
-	if (is.null(x$covmat[1][[1]]) == F) {
-		for (j in 1:K) {
-			names(obj)[j] <- paste("stage",j,sep="")
-			obj[[j]] <- matrix(nrow=length(x$psi[[j]]),ncol=2)
-			rownames(obj[[j]]) <- x$blip.list[[j]]
-			colnames(obj[[j]]) <- c(paste(100*((1 - level)/2),"%"), paste(100*((1 + level)/2),"%"))
-			for (p in 1:length(x$psi[[j]])) {
-				obj[[j]][p,] <- c(x$psi[[j]][p] + qnorm((1-level)/2) * sqrt(x$covmat[[j]][p,p]), x$psi[[j]][p] + qnorm((1+level)/2)*sqrt(x$covmat[[j]][p,p]))
-			}
-		}
-	return(obj)
-	} else {
-		cat("confint() only available when DTRreg called with a variance estimation option.\n")
-	}
+confint.DTRreg <- function(object, parm = NULL, level = 0.95, type = "se", ...) {
+  x <- object
+  K <- x$K
+  obj <- x$psi
+  if (is.null(x$covmat[1][[1]]) == F) {
+    if(type == "se"){
+      for (j in 1:K) {
+        names(obj)[j] <- paste("stage",j,sep="")
+        obj[[j]] <- matrix(nrow=length(x$psi[[j]]),ncol=2)
+        rownames(obj[[j]]) <- x$blip.list[[j]]
+        colnames(obj[[j]]) <- c(paste(100*((1 - level)/2),"%"), paste(100*((1 + level)/2),"%"))
+        for (p in 1:length(x$psi[[j]])) {
+          obj[[j]][p,] <- c(x$psi[[j]][p] + qnorm((1-level)/2) * sqrt(x$covmat[[j]][p,p]), x$psi[[j]][p] + qnorm((1+level)/2)*sqrt(x$covmat[[j]][p,p]))
+        }
+      }
+    }else if(type == "percentile"){
+      if(length(x$psi.boot) == 0){
+        stop("Percentile confidence intervals only available with bootstrap.")
+      }else{
+        for (j in 1:K) {
+          names(obj)[j] <- paste("stage",j,sep="")
+          obj[[j]] <- matrix(nrow=length(x$psi[[j]]),ncol=2)
+          rownames(obj[[j]]) <- x$blip.list[[j]]
+          colnames(obj[[j]]) <- c(paste(100*((1 - level)/2),"%"), paste(100*((1 + level)/2),"%"))
+          for (p in 1:length(x$psi[[j]])) {
+            obj[[j]][p,] <- c(quantile(x$psi.boot[[j]][,p], probs = (1-level)/2), quantile(x$psi.boot[[j]][,p], probs = (1+level)/2))
+          }
+        }
+      }
+    }
+    return(obj)
+  } else {
+    cat("confint() only available when DTRreg called with a variance estimation option.\n")
+  }
 }
 
 # diagnostics
 #' @export
-plot.DTRreg <- function(x, ...) {
+plot.DTRreg <- function(x, method = "DTRreg", ...) {
   # original Y is in x$obs.Y
   # fitted Y (one for each stage) is in x$fitted
-  cat("Y residuals (for treatment-free and blip models):\n")
-  for (j in 1:x$K) {
-    advance <- readline("Hit <Return> to see next plot:")
-    obsK <- x$obs.Y[x$remove<j]
-    fitK <- x$fitted[[j]]
-    plot(fitK,obsK-fitK,xlab="Fitted values",ylab="Residuals",main=paste("Stage",j,"Residuals vs Fitted"))
-    abline(h=0,lty=2,col=4)
-    points(lowess(fitK,obsK - fitK),type="l",col=2,lwd=2)
-    if (length(x$blip.list[[j]]) > 1) {
-      for (var in x$blip.list[[j]][2:length(x$blip.list[[j]])]) {
-        advance <- readline("Hit <Return> to see next plot:")     
-        plot(x$data[x$remove<j,which(colnames(x$data) == var)],obsK-fitK,xlab=paste(var),ylab="Residuals",main=paste("Stage",j,"Residuals vs",var))
-        abline(h=0,lty=2,col=4)
-        points(lowess(x$data[x$remove<j,which(colnames(x$data) == var)],obsK-fitK),type="l",col=2,lwd=2)
+  if(method == "DTRreg"){
+    cat("Y residuals (for treatment-free and blip models):\n")
+    for (j in 1:x$K) {
+      advance <- readline("Hit <Return> to see next plot:")
+      obsK <- x$obs.Y[x$remove<j]
+      fitK <- x$fitted[[j]]
+      plot(fitK,obsK-fitK,xlab="Fitted values",ylab="Residuals",main=paste("Stage",j,"Residuals vs Fitted"))
+      abline(h=0,lty=2,col=4)
+      points(lowess(fitK,obsK - fitK),type="l",col=2,lwd=2)
+      if (length(x$blip.list[[j]]) > 1) {
+        for (var in x$blip.list[[j]][2:length(x$blip.list[[j]])]) {
+          advance <- readline("Hit <Return> to see next plot:")
+          plot(x$data[x$remove<j,which(colnames(x$data) == var)],obsK-fitK,xlab=paste(var),ylab="Residuals",main=paste("Stage",j,"Residuals vs",var))
+          abline(h=0,lty=2,col=4)
+          points(lowess(x$data[x$remove<j,which(colnames(x$data) == var)],obsK-fitK),type="l",col=2,lwd=2)
+        }
       }
     }
-  }
-  cat("Treatment models:\n")
-  for (j in 1:x$K) {
-    cat("Stage",j,"\n")
-    plot(x$treat.mod.fitted[[j]],sub.caption=paste("Stage",j,"treatment"))
+    cat("Treatment models:\n")
+    for (j in 1:x$K) {
+      cat("Stage",j,"\n")
+      plot(x$treat.mod.fitted[[j]],sub.caption=paste("Stage",j,"treatment"))
+    }
+  }else if(method == "DWSurv"){
+    cat("Y residuals (for treatment-free and blip models):\n")
+    for (j in 1:x$K) {
+      advance <- readline("Hit <Return> to see next plot:")
+      obsK <- x$log.obs[[j]]
+      fitK <- x$log.fitted[[j]]
+      plot(fitK,obsK-fitK,xlab="Fitted values",ylab="Residuals",main=paste("Stage",j,"Residuals vs Fitted"))
+      abline(h=0,lty=2,col=4)
+      points(lowess(fitK,obsK - fitK),type="l",col=2,lwd=2)
+    }
+    cat("Treatment models:\n")
+    for (j in 1:x$K) {
+      advance <- readline("Hit <Return> to see next plot:")
+      cat("Stage",j,"\n")
+      plot(x$treat.mod.fitted[[j]],sub.caption=paste("Stage",j,"treatment"))
+    }
+    cat("Censoring models:\n")
+    for (j in 1:x$K) {
+      advance <- readline("Hit <Return> to see next plot:")
+      cat("Stage",j,"\n")
+      plot(x$cens.mod.fitted[[j]],sub.caption=paste("Stage",j,"censoring"))
+    }
   }
 }
 
@@ -674,28 +722,28 @@ get_all_vars_DTR <- function(x,n) {
 chooseM <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "gest", weight = "default", missing = "default", treat.mod.man = NULL, B1 = 500, B2 = 500)
 {
 	# fit initial DTRreg model to the data to get original estimates + phat
-	mod1 <- DTRreg(outcome,blip.mod, treat.mod, tf.mod, data, method, weight, missing, treat.mod.man, var.estim="bootstrap", M = 0, truncate = 0, B = 200)
+	mod1 <- DTRreg(outcome,blip.mod, treat.mod, tf.mod, data, method, weight, missing, treat.mod.man, var.estim="bootstrap", M = 0, truncate = 0, B = 50)
 	blip.psi.1 <- mod1$psi[[1]]
 	p <- mod1$nonreg[2]
-	
+
 	# make sure the data are in a data frame
 	if (is.null(data)) {
     data <- cbind(outcome,do.call("cbind",lapply(blip.mod,get_all_vars)),do.call("cbind",lapply(treat.mod,get_all_vars)),do.call("cbind",lapply(tf.mod,get_all_vars)))
     data <- data[!duplicated(lapply(data,summary))]
 	}
   	data <- data.frame(data)
-	
+
 	# deal with missing data here
-	
-	# initialize 
+
+	# initialize
 	alpha <- 0.025
 	n <- nrow(data)
-	
+
 	while(alpha <= 0.5) # loop until alpha reaches 0.5, with break when the coverage reaches the nominal 95% rate
 	{
 		# reset coverage
 		coverage <- 0
-		
+
 		# reset matrix to save estimates
     	est <- matrix(NA, ncol = B2 + 3, nrow = 1)
     	for(j in 1:B1) # loop over B1 first stage bootstrap samples
@@ -703,7 +751,7 @@ chooseM <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "g
      		# draw a n-out-of-n bootstrap sample
       		index <- sample(1:n, n, replace = TRUE)
       		boot1 <- as.data.frame(data[index,])
-      		
+
       		# if prespecified probability of treatment with treat.mod.man, adjust
       		treat.mod.manN <- NULL
       		if(class(treat.mod.man)=="list")
@@ -712,26 +760,26 @@ chooseM <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "g
       			numT <- length(treat.mod.man)
       			for(t in 1:numT) treat.mod.manN[[t]] <- treat.mod.man[[t]][index]
       		}
-      
+
       		# fit the model to b1-th bootstrap sample
       		res1 <- try(DTRreg(outcome, blip.mod, treat.mod, tf.mod, treat.mod.man = treat.mod.manN, method, var.estim = "bootstrap", data = boot1))
       		esb1 <- res1$psi[[1]][1] # only consider main effect of treatment
       		est[1] <- esb1
-      
+
       		# estimate m for each b1 bootstrap sample
       		phat <- res1$nonreg[2]
       		est[2] <- phat
-      
+
       		# resampling size
-      		m <- n^((1 + alpha*(1-phat))/(1 + alpha)) 
+      		m <- n^((1 + alpha*(1-phat))/(1 + alpha))
       		est[3] <- m
-      
+
            	for(k in 1:B2) # loop over B2 second stage bootstrap samples
       		{
-        		# resample with replacement 
+        		# resample with replacement
         		index <- sample(1:n, floor(m), replace = TRUE)
         		boot2 <- boot1[index,]
-        		
+
         		# if prespecified probability of treatment with treat.mod.man
         		treat.mod.manM <- NULL
       			if(class(treat.mod.man)=="list")
@@ -740,11 +788,11 @@ chooseM <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "g
       				numT <- length(treat.mod.man)
       				for(t in 1:numT) treat.mod.manM[[t]] <- treat.mod.manN[[t]][index]
       			}
-      			
+
         		# fit the model to bootstrap sample
         		res2 <- try(DTRreg(outcome, blip.mod, treat.mod, tf.mod, treat.mod.man = treat.mod.manM, method, data = as.data.frame(boot2)))
         		esb2 <- res2$psi[[1]][1]
-        
+
         		# save the (b1,b2) bootstrap estimates in the (k+3) column
         		est[k + 3] <- esb2
       		}
@@ -757,15 +805,354 @@ chooseM <- function(outcome, blip.mod, treat.mod, tf.mod, data=NULL, method = "g
 		if(coverage >= 0.95)
     	{
       		cat("Final value of alpha = ", alpha,"\n")
-      		m <- n^((1 + alpha*(1-p))/(1 + alpha)) 
+      		m <- n^((1 + alpha*(1-p))/(1 + alpha))
       		cat("Selected subsample size m = ", floor(m),"\n")
       		break
     	} else {
        		alpha <- alpha + 0.025
-    	}    	
+    	}
 	}
-	return(list(m=as.numeric(floor(m))))  
+	return(list(m=as.numeric(floor(m))))
 }
 
 
-
+# Doubly-robust dynamic treatment regimen estimation via dynamic weighted survival modeling (DWSurv), assuming linear blip and treatment-free functions and logistic regression for treatment and censoring models
+#' @export
+DWSurv <- function(time, blip.mod, treat.mod, tf.mod, cens.mod, data = NULL, weight = "default", var.estim = "none", asymp.opt = "adjusted", boot.opt = "standard", B = 500, optimization = "max", quiet = FALSE) {
+  obj <- list()
+  # for compatibility with other functions
+  obj$type <- "DTR"
+  # save formula
+  obj$time.mod <- time
+  obj$blip.mod <- blip.mod
+  obj$treat.mod <- treat.mod
+  obj$tf.mod <- tf.mod
+  obj$cens.mod <- cens.mod
+  # checking input validity
+  try(match.arg(var.estim, c("none", "asymptotic", "bootstrap")))
+  if(var.estim == "asymptotic") try(match.arg(asymp.opt, c("adjusted", "naive")))
+  if(var.estim == "bootstrap") try(match.arg(boot.opt, c("standard", "empirical", "normal")))
+  try(match.arg(optimization, c("min", "max")))
+  if (length(treat.mod) != length(blip.mod) | length(treat.mod) != length(tf.mod) | length(treat.mod) != length(cens.mod) | length(treat.mod) != length(time)) {stop("-treat.mod-, -blip.mod-, -tf.mod-, -cens.mod- and -time- must all have same length.")}
+  # infer stages from blip.mod
+  K <- length(blip.mod)
+  obj$K <- K
+  obj$optimization <- optimization
+  # checking input validity (continued)
+  for(j in 1:K){
+    if(class(treat.mod[[j]]) != "formula") stop(paste("The stage ", j, " treatment model must be supplied as a formula with the treatment variable on the left hand side.", sep = ""))
+    if(class(cens.mod[[j]]) != "formula") stop("The stage ", j, " censoring model must be supplied as a formula with the censoring indicator on the left hand side.")
+    if(class(time[[j]]) != "formula" | length(all.vars(time)) > 1) stop("The stage ", j, " survival time must be supplied as a formula with nothing on the left hand side and the survival time variable on the right hand side.")
+  }
+  #time <- list(time); treat.mod <- list(treat.mod); blip.mod <- list(blip.mod); tf.mod <- list(tf.mod); cens.mod <- list(cens.mod)
+  
+  # if no data set, create one from the models
+  if (is.null(data)) {
+    data <- cbind(do.call("cbind",lapply(time,get_all_vars)), do.call("cbind",lapply(blip.mod,get_all_vars)),do.call("cbind",lapply(treat.mod,get_all_vars)),do.call("cbind",lapply(tf.mod,get_all_vars)), do.call("cbind",lapply(cens.mod,get_all_vars)))
+    data <- data[!duplicated(lapply(data,summary))]
+    obj$time <- lapply(time,get_all_vars)
+    obj$status <- do.call("cbind",lapply(cens.mod,get_all_vars))[,1]
+  } else {
+    obj$time <- list()
+    for(j in 1:K){
+      obj$time[[j]] <- data[,which(colnames(data) == all.vars(time[[j]]))]
+    }
+    obj$status <- data[,which(colnames(data) == all.vars(cens.mod[[1]])[1])] 
+  }
+  # make sure data is a data frame (can't be a matrix)
+  data <- data.frame(data)
+  data$id <- seq(1:nrow(data))
+  obj$data <- data
+  Y <- obj$time
+  delta <- obj$status
+  
+  if (any(is.na(delta))) { # missing survival times will be normal if patients don't reach stage
+    stop("Missing values in status are not allowed.")
+  }
+  obj$blip.list <- obj$psi <- obj$covmat <- obj$covmat.all <- list() # to output list of covariates in blip functions, list of estimates of psi, var-covar matrix at each stage
+  N <- nrow(data)
+  # calculate eta j i.e. whether an individual enters stage j and remove individuals with missing data
+  eta <- list()
+  drop <- c()
+  for(j in 1:K) {
+    if(j > 1){
+      eta[[j]] <- ifelse(is.na(Y[[j]]) | Y[[j]] == 0 | eta[[j-1]] == 0, 0, 1)
+    }else{
+      eta[[j]] <- ifelse(is.na(Y[[j]]) | Y[[j]] == 0, 0, 1)
+    }
+    dataetaj <- data[eta[[j]] == 1,]
+    A <- model.response(model.frame(treat.mod[[j]], dataetaj, na.action = 'na.pass'))
+    Hpsi <- model.matrix(blip.mod[[j]], model.frame(blip.mod[[j]], dataetaj, na.action = 'na.pass'))
+    Hbeta <- model.matrix(tf.mod[[j]], model.frame(tf.mod[[j]], dataetaj, na.action = 'na.pass'))
+    Halpha <- model.matrix(treat.mod[[j]], model.frame(treat.mod[[j]], dataetaj, na.action = 'na.pass'))
+    if(is.null(cens.mod[[j]]) != TRUE) {
+      Hlambda <- model.matrix(cens.mod[[j]], model.frame(cens.mod[[j]], dataetaj, na.action = 'na.pass'))
+      drop <- append(drop,dataetaj$id[which(apply(is.na(cbind(A, Hpsi, Hbeta, Halpha, Hlambda)), 1, any))])
+    }else{
+      drop <- append(drop,dataetaj$id[which(apply(is.na(cbind(A, Hpsi, Hbeta, Halpha)), 1, any))])
+    }
+    drop <- unique(drop)
+    if(length(drop) > 0) {
+      eta[[j]][which(data$id %in% drop)] <- 0
+      Y[[j]][which(data$id %in% drop)] <- 0
+    }
+  }
+  
+  # keep track of id removed
+  obj$drop <- drop
+  # keep track of who enter stages
+  obj$eta <- eta
+  # store quantities for variance estimation
+  A.store <- Hpsi.store <- U.adj.store <- HD.store <- Htheta.store <- Halpha.store <- ds.dalpha.store <- score.alpha.store <- list()
+  delta.store <- w.store <- Ahat.store <- Hlambda.store <- ds.dlambda.store <- score.lambda.store <- Dhat.store <- list()
+  # save bootstrap estimates across stages if necessage
+  obj$psi.boot <- list()
+  # work in stages starting from final stage (K)
+  Ycumul <- rep(0, nrow(data))
+  for (j in K:1) {
+    Ycumul[eta[[j]] == 1] <- Ycumul[eta[[j]] == 1] + Y[[j]][eta[[j]] == 1]
+    Yj <- Ycumul[eta[[j]] == 1]
+    dataetaj <- data[eta[[j]] == 1, ]
+    deltaj <- delta.store[[j]] <- delta[eta[[j]] == 1]
+    
+    # get data: Hpsi = blip, Hbeta = treatment-free, Halpha = treatment, Hlambda = censoring
+    A <- A.store[[j]] <- as.numeric(model.response(model.frame(treat.mod[[j]], dataetaj, na.action = 'na.pass')))
+    Hpsi <- model.matrix(blip.mod[[j]], model.frame(blip.mod[[j]], dataetaj, na.action = 'na.pass'))
+    Hbeta <- model.matrix(tf.mod[[j]], model.frame(tf.mod[[j]], dataetaj, na.action = 'na.pass'))
+    Halpha <- model.matrix(treat.mod[[j]], model.frame(treat.mod[[j]], dataetaj, na.action = 'na.pass'))
+    if(is.null(cens.mod[[j]]) != TRUE) Hlambda <- model.matrix(cens.mod[[j]], model.frame(cens.mod[[j]], dataetaj, na.action = 'na.pass'))
+    
+    # list of censoring and treatment variables
+    treat.var <- all.vars(treat.mod[[j]])[1]
+    
+    # store blip variable names for output
+    obj$blip.list[[j]] <- colnames(Hpsi)
+    
+    # force matrices in case intercept-only models
+    Hpsi <- Hpsi.store[[j]] <-  as.matrix(Hpsi)
+    Hbeta <- as.matrix(Hbeta)
+    Halpha <- Halpha.store[[j]] <- as.matrix(Halpha)
+    if(is.null(cens.mod[[j]]) != TRUE) Hlambda <- Hlambda.store[[j]] <- as.matrix(Hlambda)
+    obj$n[[j]] <- nrow(dataetaj)
+    
+    # check if treatments are binary, if not, and dWOLS selected, abort
+    A.bin <- as.numeric(length(unique(A)) <= 2)
+    if (A.bin == 0) {
+      stop("Non-binary treatment not suitable for DWSurv analysis.\n")
+    }
+    # if binary and not 0/1, recode (and print warning)
+    if (A.bin & !all(unique(A) %in% c(0, 1))) {
+      A <- A.store[[j]] <- ifelse(A == min(A),0,1)
+      # suppress output if bootstrapping
+      if (quiet == FALSE) {
+        cat("Warning: stage", j, "treatment recoded to 0/1.\n")
+      }
+    }
+    # for compatibility with other functions
+    obj$cts[[j]] <- "bin"
+    
+    # treatment model: binary logistic
+    alpha <- glm(A ~ -1 + Halpha, family = binomial(link = "logit"))
+    obj$treat.mod.fitted[[j]] <- alpha
+    Ahat <- Ahat.store[[j]] <- fitted(alpha)
+    obj$obs.treat[[j]] <- A
+    
+    # censoring model: binary logistic
+    if(sum(deltaj) != length(deltaj) & is.null(cens.mod[[j]]) != TRUE){
+      lambda <- glm(deltaj ~ -1 + Hlambda, family = binomial(link = "logit"))
+      obj$cens.mod.fitted[[j]] <- lambda
+      Dhat <- Dhat.store[[j]] <- fitted(lambda)
+      obj$obs.cens[[j]] <- deltaj
+    }else if(sum(deltaj) == length(deltaj) & is.null(cens.mod[[j]]) != TRUE){
+      Dhat <- Dhat.store[[j]] <- deltaj
+      cat("Warning: no censoring in stage", j,", censoring model will be ignored.\n")
+    }else{
+      Dhat <- Dhat.store[[j]] <- deltaj
+    }
+    
+    # estimation step
+    # weights
+    if (is.function(weight) == FALSE) {
+      w <- w.store[[j]] <- abs(A - Ahat)/(deltaj*Dhat + (1 - deltaj)*(1 - Dhat))
+    } else {
+      # weight function specified for pi = 1 and delta = 1, therefore
+      w <- w.store[[j]] <- weight(A, Ahat, deltaj, Dhat)
+    }
+    # blip parameter estimates extracted via dimensions of Hbeta and Hpsi
+    logY <- log(Yj[deltaj == 1])
+    wj <- w[deltaj == 1]
+    Hbetaj <- Hbeta[deltaj == 1, ]
+    Hpsij <- Hpsi[deltaj == 1, ]
+    Aj <- A[deltaj == 1]
+    est <- solve(t(cbind(Hbetaj, Aj * Hpsij)) %*% cbind(wj * Hbetaj, wj * Aj * Hpsij)) %*% t(cbind(Hbetaj, Aj * Hpsij)) %*% (wj * logY)
+    psi <- est[(dim(Hbeta)[2] + 1):(dim(Hbeta)[2] + dim(Hpsi)[2])]
+    beta <- est[1:dim(Hbeta)[2]]
+    
+    # if parametric bootstrap requested
+    if(var.estim == "bootstrap" & boot.opt != "standard"){
+      psi.boot <- matrix(NA, nrow = B, ncol = length(psi))
+      residuals <- logY - Hbetaj %*% beta - Aj * Hpsij %*% psi
+      reshist <- hist(residuals, plot = FALSE)
+      mean.res <- mean(residuals)
+      sd.res <- sd(residuals)
+      boot.data <- dataetaj
+      for (k in 1:B) {
+        boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])] <- rbinom(obj$n[[j]], 1, Dhat.store[[j]])
+        if(boot.opt == "empirical"){
+          bins <- with(reshist, sample(length(mids), sum(boot.data[,which(colnames(boot.data) == all.vars(cens.mod[[1]])[1])]), prob = density, replace = TRUE))
+          newres <- runif(obj$n[[j]], reshist$breaks[bins], reshist$breaks[bins + 1])
+        }else if(boot.opt == "normal"){
+          newres <- rnorm(obj$n[[j]], mean = mean.res, sd = sd.res)
+        }
+        boot.data[, which(colnames(boot.data) == all.vars(time[[j]]))] <- exp(Hbeta %*% beta + A*Hpsi %*% psi + newres)
+        psi.boot[k,] <- DWSurv(time = list(obj$time.mod[[j]]), blip.mod = list(obj$blip.mod[[j]]), treat.mod = list(obj$treat.mod[[j]]), tf.mod = list(obj$tf.mod[[j]]), cens.mod = list(obj$cens.mod[[j]]), boot.data, var.estim="none", quiet = TRUE)$psi[[1]]
+      }
+      obj$covmat[[j]] <- var(psi.boot)
+      obj$psi.boot[[j]] <- psi.boot
+    }
+    
+    # use estimates to identify optimal treatments
+    # depending on optimization goal
+    if (optimization == "max") {
+      opt <- as.numeric(Hpsi %*% psi > 0)
+    }else{
+      opt <- as.numeric(Hpsi %*% psi < 0)
+    }
+    # ensure opt is in vector form
+    opt <- as.vector(opt)
+    # store estimates, optimal treatments and blip parameters
+    obj$beta[[j]] <- beta
+    obj$psi[[j]] <- psi
+    obj$opt.treat[[j]] <- opt
+    # store fitted values
+    obj$log.fitted[[j]] <- Hbetaj %*% beta + Hpsij %*% psi
+    obj$log.obs[[j]] <- logY
+    # update pseudo-Y by adding regret for everybody who enter stage j
+    obj$log.regret[[j]] <- (opt - A)*(Hpsi %*% psi)
+    Yj <- exp(log(Yj) + obj$log.regret[[j]])
+    old.Ycumul <- Ycumul
+    Ycumul[eta[[j]] == 1] <- Yj
+    
+    # Asymptotic variance estimator done stage-by-stage
+    if (var.estim == "asymptotic" & is.function(weight) == FALSE) {
+      logY <- as.matrix(log(old.Ycumul[eta[[j]] == 1]))
+      theta <- as.matrix(est)
+      Htheta <- Htheta.store[[j]] <- as.matrix(cbind(Hbeta, A*Hpsi))
+      # residuals are the (Y-gamma-beta) residuals
+      residuals <- as.vector(logY - Htheta %*% theta)
+      # HD: what precedes the (Y-gamma-beta) term in the est eq.
+      HD <- HD.store[[j]] <- deltaj * w * Htheta
+      # estimating equation = (Y - gamma - E[Y - gamma|H])*(A - Ahat)*Hpsi
+      U <- residuals * HD
+      U.adj <- U.adj.store[[j]] <- U
+      Ahat <- as.vector(Ahat)
+      Dhat <- as.vector(Dhat)
+      A <- as.vector(A)
+      nj <- sum(eta[[j]])
+      # term that depends on alpha treatment model
+      Halpha <- as.matrix(Halpha)
+      score.alpha <- score.alpha.store[[j]] <- (A - Ahat) * Halpha
+      ds.dalpha <- ds.dalpha.store[[j]] <- (-1/nj) * (crossprod(Ahat * (1 - Ahat) * Halpha, Halpha))
+      dU.dalpha <- (1/nj) * (crossprod(deltaj * abs(A - Ahat)/(deltaj * Dhat + (1 - deltaj) * (1 - Dhat)) * (1 - 2*A) * Ahat * (1 - Ahat) * residuals * Htheta, Halpha))
+      U.adj <- U.adj.store[[j]] <- U.adj - t(dU.dalpha %*% solve(ds.dalpha) %*% t(score.alpha))
+      # terms that depends on lambda censoring model
+      if(is.null(cens.mod[[j]]) != TRUE){
+        Hlambda <- as.matrix(Hlambda)
+        score.lambda <- score.lambda.store[[j]] <- (deltaj - Dhat) * Hlambda
+        ds.dlambda <- ds.dlambda.store[[j]] <- (-1/nj) * (crossprod(Dhat * (1 - Dhat) * Hlambda, Hlambda))
+        dU.dlambda <- (-1/nj) * (crossprod(deltaj * abs(A - Ahat)/(deltaj * Dhat + (1 - deltaj) * (1 - Dhat))^2 * (2*deltaj - 1) * Dhat * (1 - Dhat) * residuals * Htheta, Hlambda))
+        U.adj <- U.adj.store[[j]] <- U.adj - t(dU.dlambda %*% solve(ds.dlambda) %*% t(score.lambda))
+      }
+      # add the terms that depend on previously estimated parameters
+      if(j < K){
+        for(l in (j+1):K){
+          Hpsitemp <- matrix(0, nrow = N, ncol = ncol(Hpsi.store[[l]]))
+          Hpsitemp[eta[[l]] == 1,] <- Hpsi.store[[l]]
+          psitemp <- as.vector(obj$psi[[l]])
+          px <- Hpsitemp %*% psitemp
+          px <- px[eta[[j]] == 1]
+          y <- rep(0, N)
+          y[eta[[l]] == 1] <- Y[[l]][eta[[l]] == 1]
+          y <- y[eta[[j]] == 1] # survival time in interval l - ensure coded with 0 if didnt enter stage l
+          optl <- rep(0, N)
+          optl[eta[[l]] == 1] <- obj$opt.treat[[l]] - A.store[[l]]
+          optl <- optl[eta[[j]] == 1]
+          etal <- eta[[l]][eta[[j]] == 1]
+          B <- (Ycumul[eta[[j]] == 1])^(-1) * etal * y * exp(px * (optl)) * (optl)
+          dUj.dpsil <- (1/nj)*crossprod(as.vector(deltaj * w * B) * Hpsitemp[eta[[j]] == 1,], Htheta)
+          Hpsitemp <- Hpsitemp[eta[[l]] == 1, ]
+          njp1 <- obj$n[[l]]
+          dUjp1.dpsijp1 <- (-1/njp1) * (crossprod(HD.store[[l]], Hpsitemp))
+          # treatment component
+          dUjp1.dpsijp1 <- dUjp1.dpsijp1 - t(Htheta.store[[l]]) %*% Halpha.store[[l]] %*% solve(ds.dalpha.store[[l]]) %*% t(score.alpha.store[[l]]) %*% (-1/njp1 * delta.store[[l]] * w.store[[l]] * (1 - 2 * A.store[[l]]) * Ahat.store[[l]] * (1 - Ahat.store[[l]]) * Hpsitemp)
+          # censoring component if necessary
+          if(is.null(cens.mod[[j]]) != TRUE){
+            dUjp1.dpsijp1 <- dUjp1.dpsijp1 - t(Htheta.store[[l]]) %*% Hlambda.store[[l]] %*% solve(ds.dlambda.store[[l]]) %*% t(score.lambda.store[[l]]) %*% (1/njp1 * delta.store[[l]] * abs(A.store[[l]] - Ahat.store[[l]])/(delta.store[[l]] * Dhat.store[[l]] + (1 - delta.store[[l]]) * (1 - Dhat.store[[l]]))^2 * (2 * delta.store[[l]] - 1) * Dhat.store[[l]] * (1 - Dhat.store[[l]]) * Hpsitemp)
+          }
+          dUjp1.dpsijp1 <- dUjp1.dpsijp1[(length(obj$beta[[l]]) + 1):(length(obj$beta[[l]]) + length(obj$psi[[l]])),]
+          U.adje <- matrix(0, ncol = length(obj$psi[[l]]), nrow = N)
+          U.adje[eta[[l]] == 1,] <- U.adj.store[[l]][,(length(obj$beta[[l]]) + 1):(length(obj$beta[[l]]) + length(obj$psi[[l]]))]
+          U.adje <- U.adje[eta[[j]] == 1,]
+          U.adj <- U.adj - t(t(dUj.dpsil) %*% solve(dUjp1.dpsijp1) %*% t(U.adje))
+        }
+      }
+      # derivative with respect to parameter of interest
+      dUadj.dtheta <- dU.dtheta <- (-1/nj) * (crossprod(HD, Htheta))
+      if(asymp.opt == "adjusted"){
+        dUadj.dtheta <- dUadj.dtheta  - t(Htheta) %*% Halpha %*% solve(ds.dalpha) %*% t(score.alpha) %*% (-1/nj * deltaj * w * (1 - 2 * A) * Ahat * (1 - Ahat) * Htheta)
+        if(is.null(cens.mod[[j]]) != TRUE){
+          dUadj.dtheta <- dUadj.dtheta - t(Htheta) %*% Hlambda %*% solve(ds.dlambda) %*% t(score.lambda) %*% (1/nj * deltaj * abs(A - Ahat)/(deltaj * Dhat + (1 - deltaj) * (1 - Dhat))^2 * (2 * deltaj - 1) * Dhat * (1 - Dhat) * Htheta)
+        }
+        invdUadj.dtheta <- solve(dUadj.dtheta)
+        IF <- U.adj %*% invdUadj.dtheta
+        sigma.theta <- (1/nj) * var(IF)
+        obj$covmat.all[[j]] <- sigma.theta
+        obj$covmat[[j]] <- obj$covmat.all[[j]][(length(obj$beta[[j]]) + 1):(length(obj$beta[[j]]) + length(obj$psi[[j]])), (length(obj$beta[[j]]) + 1):(length(obj$beta[[j]]) + length(obj$psi[[j]]))]
+      }else{
+        invdU.dtheta <- solve(dU.dtheta)
+        IF <- U %*% invdU.dtheta
+        sigma.theta <- (1/nj) * var(IF)
+        obj$covmat.all[[j]] <- sigma.theta
+        obj$covmat[[j]] <- obj$covmat.all[[j]][(length(obj$beta[[j]]) + 1):(length(obj$beta[[j]]) + length(obj$psi[[j]])), (length(obj$beta[[j]]) + 1):(length(obj$beta[[j]]) + length(obj$psi[[j]]))]
+      }
+    }
+  }
+  
+  # optimal outcome relevant only for observed events
+  obj$Y.opt <- Ycumul
+  
+  # standard errors
+  # standard bootstrap
+  if (var.estim == "bootstrap" & boot.opt == "standard") {
+    psi.boot <- list()
+    M <- nrow(data)
+    for (i in 1:B) {
+      data.boot <- data[sample(1:M, replace=TRUE),]
+      psi.boot[[i]] <- DWSurv(time = obj$time.mod, blip.mod = obj$blip.mod, treat.mod = obj$treat.mod, tf.mod = obj$tf.mod, cens.mod = obj$cens.mod, data.boot, var.estim="none", quiet = TRUE)$psi
+    }
+    psi.boot <- do.call(function(...) mapply(rbind, ..., SIMPLIFY=FALSE), psi.boot)
+    covmat <- lapply(psi.boot, var)
+    obj$covmat <- covmat
+    obj$psi.boot <- psi.boot
+  }
+  
+  # if variance requested, calculate non-regularity
+  if (var.estim != "none" & A.bin == 1) {
+    for (j in K:1) {
+      Hpsi <- model.matrix(blip.mod[[j]],data)
+      # lower/upper limits on psi
+      psi.l <- obj$psi[[j]] - 1.96*sqrt(diag(obj$covmat[[j]]))
+      psi.u <- obj$psi[[j]] + 1.96*sqrt(diag(obj$covmat[[j]]))
+      # look at max/min value of blip based on sign of covariates and max/min of parameter CIs
+      Hpsi.p <- apply(Hpsi,c(1,2),FUN=function(x) max(x,0))
+      Hpsi.n <- apply(Hpsi,c(1,2),FUN=function(x) min(x,0))
+      # lower blip is positive values * lower CI + negative values * upper CI
+      blip.l <- Hpsi.p %*% psi.l + Hpsi.n %*% psi.u
+      blip.u <- Hpsi.p %*% psi.u + Hpsi.n %*% psi.l
+      obj$nonreg[j] <- sum(blip.l < 0 & blip.u > 0)/dim(Hpsi)[1]
+    }
+  }
+  # return
+  class(obj) <- "DTRreg"
+  obj
+}
